import os
import re
from typing import Any, Dict, List, Optional, Tuple, Union

from bs4 import BeautifulSoup, PageElement

PDF_LOCAL_OPTIONS: List[Tuple[str, Union[bool, str, None]]] = [
    ("pdf-build", True),
    ("pdf-title", None),
    ("pdf-subtitle", []),
    ("pdf-type", "document"),
    ("pdf-filename", None),
    ("pdf-revision", None),
]


def get_pdf_metadata(pdf_metadata: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    """
    Function for parsing RST file metadata. The metadata includes the ff.:

    - pdf-build: default to True
    - pdf-title: default to None
    - pdf-subtitle: default to empty list []. If a value is provided then the list must be separated by the ``|`` symbol
    - pdf-type: default to "document"
    - pdf-filename: default to None
    - pdf-revision: default to None
    - pdf-toc_txt: default to False

    :param pdf_metadata: A dictionary object containing the metadata of a page.
    :return: Dictionary object containing parsed metadata.
    """

    pdf_metadata = {} if pdf_metadata is None else pdf_metadata
    pdf_meta: Dict[str, Any] = {}

    for option, default_value in PDF_LOCAL_OPTIONS:
        option_value = default_value
        if option in pdf_metadata:
            option_value = pdf_metadata.get(option)
            if option == "pdf-build":
                option_value = option_value.lower() != "false"
            elif option == "pdf-toc_txt":
                option_value = option_value.lower() == "true"
            elif option == "pdf-subtitle":
                option_value = [item.strip() for item in option_value.split("|")]

        pdf_meta.update([(option.replace("pdf-", "", 1), option_value)])
    return pdf_meta


def secure_filename(filename):
    r"""Pass it a filename, and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.

    On Windows systems the function also makes sure that the file is not
    named after one of the special device files.

    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename(u'i contain cool \xfcml\xe4uts.txt')
    'i_contain_cool_umlauts.txt'

    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you generate random
    filename if the function returned an empty one.

    :param filename: the filename to secure
    """
    _filename_ascii_strip_re = re.compile(r"[^A-Za-z0-9_.-]")
    _windows_device_files = (
        "CON",
        "AUX",
        "COM1",
        "COM2",
        "COM3",
        "COM4",
        "LPT1",
        "LPT2",
        "LPT3",
        "PRN",
        "NUL",
    )

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, " ")
    filename = str(_filename_ascii_strip_re.sub("", "_".join(filename.split()))).strip("._")

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if os.name == "nt" and filename and filename.split(".")[0].upper() in _windows_device_files:
        filename = "_" + filename

    return filename


def h1_title_tag(content: Union[str, PageElement], pdf_title: str) -> Optional[str]:
    soup = content
    if isinstance(soup, str):
        soup = BeautifulSoup(soup, "html5lib")
    title = soup.find("h1", attrs={"id": re.compile(r"[\w\-]+")})
    if title is None:
        return pdf_title
    title = re.sub(r"^[\d.]+ ", "", title.text)
    return title.rstrip("¶")



from pathlib import Path
from urllib.parse import urlparse

from . import _FilterBase


class URLFilter(_FilterBase):
    """Finds a matching filename in some directories and returns its URL."""

    def __call__(self, pathname: str) -> str:
        if not pathname:
            return ""

        # Check for URL(eg. 'https://...')
        target_url = urlparse(pathname)
        if target_url.scheme or target_url.netloc:
            return pathname

        # Search image file in below directories:
        dirs = [
            Path(self.config["outdir"]).resolve(),
            Path(self.config["srcdir"]).resolve(),
            Path("."),
        ]

        for d in dirs:
            if not d:
                continue
            path = Path(Path.joinpath(d, pathname)).resolve()
            if path.is_file():
                return path.as_uri()
        # return path

        # not found?
        return pathname



import logging
import re
import sys
from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import Any, Dict, Optional

from bs4 import BeautifulSoup, Tag
from weasyprint import HTML

from . import cover, toc
from .options import Options
from .preprocessor import get_content
from .preprocessor import get_separate as prep_separate
from .styles import style_for_print
from .templates.filters.url import URLFilter
from .themes import generic as generic_theme


class Renderer:
    def __init__(self, options: Options, config: Dict[str, Any]):
        self._options = options
        self._config = config

        self.theme = self._load_theme_handler()
        self.user_plugin = self._load_user_plugin_handler()
        self.pgnum = 0
        self.pages = []

    def write_pdf(
        self,
        content: str,
        base_url: str,
        filename: str,
        pdf_metadata: Optional[Dict] = None,
    ):
        self.render_doc(content, base_url, pdf_metadata=pdf_metadata).write_pdf(filename)

    def render_doc(self, content: str, base_url: str, pdf_metadata: Dict = None):
        soup = BeautifulSoup(content, "html5lib")
        soup = get_content(soup)
        self.inject_pgnum(soup)

        style_tags: list[Tag] = style_for_print(self._options, pdf_metadata)
        style_tags[0].append(self.theme.get_stylesheet())  # Add theme CSS

        for style_tag in style_tags:
            soup.head.append(style_tag)

        soup = prep_separate(soup, base_url, self._options.site_url, self._config["outdir"])
        toc.make_toc(soup, self._options)
        cover.make_cover(soup, self._options, self._config, pdf_metadata=pdf_metadata)
        if self.user_plugin:
            self.user_plugin.main(soup=soup)

        # Enable Debugging
        not_as_uri = re.compile(r"^file:/{,2}")
        pattern = r"^{}".format(self._config["outdir"].rstrip("/"))
        check_site_dir = re.compile(pattern)
        if self._options.debug and self._options.debug_target is not None:
            # Debug a single PDF build file
            path_filter = URLFilter(self._options, self._config)
            debug_target_file = path_filter(pathname=str(self._options.debug_target))
            doc_src_path = path_filter(pathname=str(self._options.rst_src_path))

            if doc_src_path == debug_target_file:
                debug_folder_path = str(self._options.debug_dir()).replace("\\", "/")
                rel_url = not_as_uri.sub("", base_url)
                pdf_html_file = check_site_dir.sub(debug_folder_path, rel_url) + ".html"
                pdf_html_dir = Path(pdf_html_file).parent
                if not pdf_html_dir.is_dir():
                    pdf_html_dir.mkdir(parents=True, exist_ok=True)
                with open(pdf_html_file, "w", encoding="UTF-8") as f:
                    f.write(soup.prettify())
                html = HTML(string=str(soup)).render()
                return html
        elif self._options.debug and self._options.debug_target is None:
            # Debug every PDF build file
            debug_folder_path = str(self._options.debug_dir()).replace("\\", "/")
            rel_url = not_as_uri.sub("", base_url)
            pdf_html_file = check_site_dir.sub(debug_folder_path, rel_url) + ".html"
            pdf_html_dir = Path(pdf_html_file).parent
            if not pdf_html_dir.is_dir():
                pdf_html_dir.mkdir(parents=True, exist_ok=True)
            with open(pdf_html_file, "w", encoding="UTF-8") as f:
                f.write(soup.prettify())

        html = HTML(string=str(soup)).render()
        return html

    def add_link(self, content: str, file_name: str = None):
        return self.theme.modify_html(content, file_name)

    def inject_pgnum(self, soup):
        pgnum_counter = soup.new_tag("style")
        pgnum_counter.string = """
        @page :first {{
            counter-reset: __pgnum__ {};
        }}
        @page {{
            counter-increment: __pgnum__;
        }}
        """.format(
            self.pgnum
        )

        soup.head.append(pgnum_counter)

    @property
    def logger(self) -> logging:
        return self._options.logger

    def _load_theme_handler(self):
        theme = self._options.theme_name
        custom_theme_handler_path = self._options.theme_handler_path
        module_name = "." + (theme or "generic").replace("-", "_")

        if custom_theme_handler_path:
            try:
                spec = spec_from_file_location(module_name, Path.cwd().joinpath(custom_theme_handler_path))
                mod = module_from_spec(spec)
                spec.loader.exec_module(mod)
                return mod
            except FileNotFoundError as e:
                self.logger.error(
                    'Could not load theme handler {} from custom directory "{}": {}'.format(
                        theme, custom_theme_handler_path, e
                    ),
                    file=sys.stderr,
                )
                pass
        try:
            return import_module(module_name, "sphinx_pdf_generate.themes")
        except ImportError as e:
            self.logger.error(
                f"Could not load theme handler {theme} because it is not a supported theme by default: {e}",
                file=sys.stderr,
            )
            return generic_theme

    def _load_user_plugin_handler(self):
        custom_plugin_handler_path = self._options.user_plugin_handler_path
        module_name = ".user_plugin"

        if custom_plugin_handler_path:
            try:
                spec = spec_from_file_location(module_name, Path.cwd().joinpath(custom_plugin_handler_path))
                mod = module_from_spec(spec)
                spec.loader.exec_module(mod)
                return mod
            except FileNotFoundError as e:
                self.logger.error(
                    'Could not load the user plugin handler from custom directory "{}": {}'.format(
                        custom_plugin_handler_path, e
                    ),
                    file=sys.stderr,
                )
                pass



import re

from bs4 import BeautifulSoup

from .content import restructure_tabbed_content
from .links import rel_html_href, replace_asset_hrefs


def get_separate(soup: BeautifulSoup, base_url: str, site_url: str, outdir: str):
    """
    Function to restructure HTML content, so we can use it to generate separate PDFs.

    :param soup: HTML content
    :param base_url: File URI of PDF file destination. e.g.("file:///home/user-1/docs/sample")
    :param site_url: Documentation site URL
    :param outdir: Sphinx documentation build file directory. e.g. _build or build
    :return: Restructured HTML content
    """
    # Transforms all relative hrefs pointing to other html docs into full site URL hrefs
    for a in soup.find_all("a", href=True):
        a["href"] = rel_html_href(base_url, a["href"], site_url, outdir)

    soup = replace_asset_hrefs(soup, base_url)  # Replace hrefs and srcs in link and img tags respectively
    soup = restructure_tabbed_content(soup)  # Restructure tabbed content in HTML
    return soup


def get_content(soup: BeautifulSoup) -> BeautifulSoup:
    """
    Function to restructure HTML content by removing all unwanted parts and leaving only the content that will be
    used in converting the PDF.

    :param soup: HTML content
    :return: Restructured HTML content
    """
    # support for sphinx-material & sphinx-immaterial theme
    content = soup.find("article", attrs={"class": "md-content__inner"})
    if content is None:
        # support for all sphinx themes
        content = soup.find(["div", "article"], attrs={"role": "main"})
    new_content = [content]
    soup.body.clear()
    soup.body.extend(new_content)
    # Check image alignment
    all_images = soup.find_all("img", attrs={"align": re.compile(r"left|right")})
    for img in all_images:
        # Modify <img> tags
        position = img["align"]
        img["style"] = f"float:{position};"
        del img["align"]
    # Check table alignment
    all_table_th = soup.find_all("th", attrs={"align": re.compile(r"left|right|center")})
    for th in all_table_th:
        # Modify <th> tags
        position = th["align"]
        th["style"] = f"text-align:{position};"
        del th["align"]
    all_table_td = soup.find_all("td", attrs={"align": re.compile(r"left|right|center")})
    for td in all_table_td:
        # Modify <td> tags
        position = td["align"]
        td["style"] = f"text-align:{position};"
        del td["align"]
    return soup



import os
import re
from pathlib import Path

from bs4 import BeautifulSoup
from weasyprint import urls


# check if href is relative --
# if it is relative it *should* be an HTML that generates a PDF doc
def is_doc(href: str):
    tail = Path(href).name
    ext = Path(tail).suffix

    absurl = urls.url_is_absolute(href)
    abspath = os.path.isabs(href)
    htmlfile = ext.startswith(".html")
    relative_link = re.search(r"^\.{,2}?[\w\-.~$&+,/:;=?@%#*]*?$", href)
    if relative_link is not None:
        return True
    if absurl or abspath or not htmlfile:
        return False

    return True


def rel_html_href(base_url: str, href: str, site_url: str, outdir: str):
    new_base_url = os.path.dirname(base_url)
    rel_url = new_base_url.replace("file://", "")
    pattern = r"^(/tmp|tmp)/pages[\w\-]+|^[\w\-:\\]+\\+(temp|Temp)\\+pages[\w\-]+|^{}".format(outdir.rstrip("/"))
    replace_build_dir = re.compile(pattern)

    internal = href.startswith("#")
    web_url = re.search(r"^(https://|http://|mailto:|tel:)", href)
    if web_url or internal or not is_doc(href):
        return href

    abs_html_href = Path(rel_url).joinpath(href).resolve()
    abs_html_href = replace_build_dir.sub(site_url.rstrip("/"), str(abs_html_href))
    abs_html_href = abs_html_href.replace("\\", "/")

    if abs_html_href:
        return urls.iri_to_uri(abs_html_href)
    return href


def abs_asset_href(href: str, base_url: str):
    if urls.url_is_absolute(href) or Path(href).is_absolute():
        return href
    return urls.iri_to_uri(urls.urljoin(base_url, href))


#
def replace_asset_hrefs(soup: BeautifulSoup, base_url: str):
    """Replace hrefs and srcs in link and img tags respectively by making all relative asset links absolute"""
    for link in soup.find_all("link", href=True):
        link["href"] = abs_asset_href(link["href"], base_url)

    for asset in soup.find_all(src=True):
        asset["src"] = abs_asset_href(asset["src"], base_url)
    return soup



from typing import Optional

from bs4 import BeautifulSoup


def get_stylesheet() -> Optional[str]:
    return ""


def modify_html(html: str, href: str) -> str:
    soup = BeautifulSoup(html, "html5lib")

    if soup.head:
        link = soup.new_tag(
            "link",
            href=href,
            rel="alternate",
            title="PDF Export",
            type="application/pdf",
        )
        soup.head.append(link)

    return str(soup)



import html
import re
from pathlib import Path
from typing import Any, Dict, Optional

from bs4 import Tag

from ..options import Options


def _css_escape(text: Optional[str]) -> str:
    """@see https://developer.mozilla.org/en-US/docs/Web/CSS/string"""

    if not text:
        return ""

    text = html.unescape(str(text))

    # -- probably not needed.
    # text = text.encode('unicode-escape').decode('ascii').replace('\\u', '\\')

    return text.replace("'", "\\27")


def style_for_print(options: Options, pdf_metadata: Optional[Dict[str, Any]] = None) -> list[Tag]:
    base_path = Path(Path(__file__).parent).resolve()
    pdf_metadata = {} if pdf_metadata is None else pdf_metadata

    css_string = """
    :root {{
        --author: '{}';
        --author-logo: url('{}');
        --copyright: '{}';
        --title: '{}';
        --subtitle: '{}';
        --type: '{}';
        --revision: '{}';
        --filename: '{}';
        --site-url: '{}';
        --base-page-orientation: a4 portrait;
        --rotated-page-orientation: a4 landscape;

    }}""".format(
        _css_escape(options.author),
        _css_escape(options.author_logo),
        _css_escape(options.copyright),
        _css_escape(pdf_metadata.get("title", options.body_title or options.cover_title)),
        _css_escape(pdf_metadata.get("subtitle", options.cover_subtitle)),
        _css_escape(pdf_metadata.get("type", "document")),
        _css_escape(pdf_metadata.get("revision", "")),
        _css_escape(pdf_metadata.get("filename", "")),
        _css_escape(re.sub(r"http://|https://", "", options.site_url)),
    )
    css_tag = Tag(name="style", attrs={"class": "plugin-default-css"})
    css_tag.append(css_string)
    css_files = ["_paging.css"]

    if options.toc:
        css_files.append("toc.css")

    if options.cover:
        css_files.append("cover.css")

    docs_src_dir = Path(options.srcdir).resolve()
    custom_css_path = Path(options.custom_css_path) if options.custom_css_path is not None else Path()
    # Add plugin custom CSS
    if options.custom_css_path is not None:
        if not custom_css_path.is_absolute():
            custom_css_path = docs_src_dir.joinpath(options.custom_css_path)
        if custom_css_path.is_dir():
            css_files.append("pdf_custom.css")

    css_styles_list: list[Tag] = []
    for css_file in css_files:
        filename = base_path.joinpath(css_file) if css_file != "pdf_custom.css" else custom_css_path.joinpath(css_file)
        if filename.is_file():
            with open(filename, encoding="UTF-8") as f:
                css_rules = f.read()
                if css_file in ["_styles.css", "_paging.css"]:
                    css_tag.append(css_rules)
                else:
                    style_tag = Tag(
                        name="style",
                        attrs={"class": "plugin-{}".format(css_file.replace(".", "-"))},
                    )
                    style_tag.append(css_rules)
                    css_styles_list.append(style_tag)

    css_styles_list.insert(0, css_tag)  # Insert default CSS tag at the start
    return css_styles_list



import re

from bs4 import BeautifulSoup


def restructure_tabbed_content(soup: BeautifulSoup) -> BeautifulSoup:
    """
    Function to restructure tabbed content in HTML, so it can be compatible for PDF conversion.

    :param soup: HTML content
    :return: Restructured HTML content
    """
    labels = soup.find_all("label", attrs={"for": re.compile(r"^__tabbed_[\d_]+$|^sd-tab-item-\d+$")})
    blocks = soup.find_all("div", attrs={"class": re.compile(r"tabbed-block|sd-tab-content")})
    tab_content = soup.find_all("div", attrs={"class": "tabbed-content"})

    new_blocks = []

    for div in blocks:
        div["class"] = "new_tabbed_block"
        new_blocks.append(div)

    for label, div in zip(labels, new_blocks):
        label.insert_after(div)

    for tab_ in tab_content:
        tab_.decompose()

    return soup



import json
import logging
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Dict

from docutils import nodes
from sphinx.application import Sphinx
from sphinx.util import docutils

from sphinx_pdf_generate.build import show
from sphinx_pdf_generate.logging import get_logger
from sphinx_pdf_generate.options import Options
from sphinx_pdf_generate.renderer import Renderer
from sphinx_pdf_generate.templates.filters.url import URLFilter
from sphinx_pdf_generate.utils import get_pdf_metadata, h1_title_tag, secure_filename
from sphinx_pdf_generate.version import __version__


def setup(app: Sphinx) -> Dict[str, Any]:
    ########################################################################
    # CONFIG_VALUES
    ######################################################################
    # Define config values
    app.add_config_value("pdfgen_verbose", False, "html", types=[bool])
    app.add_config_value("pdfgen_site_url", "http://127.0.0.1:8000", "html", types=[str])
    app.add_config_value("pdfgen_debug", False, "html", types=[bool])
    app.add_config_value("pdfgen_debug_target", "", "html", types=[str])
    app.add_config_value("pdfgen_author", None, "html", types=[str])
    app.add_config_value("pdfgen_author_logo", None, "html", types=[str])
    app.add_config_value("pdfgen_copyright", None, "html", types=[str])
    app.add_config_value("pdfgen_disclaimer", None, "html", types=[str])
    app.add_config_value("pdfgen_cover", True, "html", types=[bool])
    app.add_config_value("pdfgen_cover_title", None, "html", types=[str])
    app.add_config_value("pdfgen_cover_subtitle", None, "html", types=[str])
    app.add_config_value("pdfgen_custom_template_path", "_templates", "html", types=[str])
    app.add_config_value("pdfgen_theme_handler_path", None, "html", types=[str])
    app.add_config_value("pdfgen_plugin_handler_path", None, "html", types=[str])
    app.add_config_value("pdfgen_custom_css_path", None, "html", types=[str])
    app.add_config_value("pdfgen_toc", True, "html", types=[bool])
    app.add_config_value("pdfgen_toc_numbering", True, "html", types=[bool])
    app.add_config_value("pdfgen_toc_title", "Table of Contents", "html", types=[str])
    app.add_config_value("pdfgen_toc_level", 4, "html", types=[int])
    app.add_config_value("pdfgen_cover_images", None, "html", types=[dict])

    ######################################################################
    # ROLES
    ######################################################################
    app.add_role("pagebreak", page_break_role)
    ########################################################################
    # EVENTS
    ########################################################################
    # Make connections to events
    app.connect("builder-inited", builder_inited)
    app.connect("html-page-context", generate_sources_to_convert)
    app.connect("build-finished", build_finished)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


# ----- SPHINX-EVENTS FUNCTIONS ----- #
def builder_inited(app: Sphinx) -> None:
    if not hasattr(app.env, "sphinx_pdfgen_data"):
        # Used to store the data about of HTML files that needs to be converted.
        app.env.sphinx_pdfgen_data = {}


def generate_sources_to_convert(
    app: Sphinx, pagename: str, templatename: str, context: Dict, doctree: docutils.nodes.Node
):
    project_metadata: Dict[str, str] = app.env.metadata.get(pagename)  # Get metadata of a page
    app.env.sphinx_pdfgen_data.update([(pagename, project_metadata)])


def page_break_role(name, rawtext, text, lineno, inliner, options={}, content=[]):  # noqa B006
    """Sphinx role to insert a page break in the HTML output."""
    return [nodes.raw("", '<p class="page-break" style="margin: 0"></p>', format="html")], []


def build_finished(app: Sphinx, exception: Exception):
    global_options = {}
    global_options.update(
        verbose=app.config.pdfgen_verbose,
        site_url=app.config.pdfgen_site_url,
        debug=app.config.pdfgen_debug,
        debug_target=None if len(app.config.pdfgen_debug_target) == 0 else app.config.pdfgen_debug_target,
        author=app.config.author if not app.config.pdfgen_author else app.config.pdfgen_author,
        author_logo=app.config.html_logo if not app.config.pdfgen_author_logo else app.config.pdfgen_author_logo,
        copyright=app.config.copyright if not app.config.pdfgen_copyright else app.config.pdfgen_copyright,
        disclaimer=app.config.pdfgen_disclaimer,
        cover=app.config.pdfgen_cover,
        cover_title=app.config.pdfgen_cover_title if app.config.pdfgen_cover_title else app.config.project,
        cover_subtitle=app.config.pdfgen_cover_subtitle,
        custom_template_path=app.config.pdfgen_custom_template_path,
        theme_handler_path=app.config.pdfgen_theme_handler_path,
        plugin_handler_path=app.config.pdfgen_plugin_handler_path,
        custom_css_path=app.config.pdfgen_custom_css_path,
        toc=app.config.pdfgen_toc,
        toc_numbering=app.config.pdfgen_toc_numbering,
        toc_title=app.config.pdfgen_toc_title,
        toc_level=app.config.pdfgen_toc_level,
        cover_images=app.config.pdfgen_cover_images,
        theme_name=app.config.html_theme,
        templates_path=app.config.templates_path,
    )

    local_options = app.env.sphinx_pdfgen_data
    if "genindex" in local_options:
        local_options["genindex"] = {"pdf-build": "False"}
    if "search" in local_options:
        local_options["search"] = {"pdf-build": "False"}
    pdf_metadata = {"GLOBAL_OPTIONS": global_options, "LOCAL_OPTIONS": local_options}

    path_to_save_metadata = Path(app.outdir).joinpath("pdf_metadata.json")
    with open(path_to_save_metadata, "w") as json_file:
        json.dump(pdf_metadata, json_file, indent=4)


# ----- PDF-GENERATE-PLUGIN CLASS ----- #
class PdfGeneratePlugin:
    def __init__(self):
        self._options = None
        self._config = None
        self._logger = get_logger("sphinx-pdf-generate")
        self._logger.setLevel(logging.INFO)
        self.renderer = None
        self.generate_txt = None
        self.combined = False
        self.pdf_num_files = 0
        self.txt_num_files = 0
        self.num_errors = 0
        self.total_time = 0

    def on_config(self, config):
        self._config = config

        if self._config.get("debug"):
            self._logger.info("PDF debug option is enabled.")
        if self._config.get("debug_target"):
            self._logger.info("Debug Target File: {}".format(self._config.get("debug_target")))

        self._options = Options(self._config, self._logger)

        from weasyprint.logger import LOGGER

        if self._options.verbose:
            LOGGER.setLevel(logging.DEBUG)
            self._logger.setLevel(logging.DEBUG)
        else:
            LOGGER.setLevel(logging.ERROR)

        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
        LOGGER.addHandler(handler)

        self.renderer = Renderer(options=self._options, config=self._config)
        return

    def convert_page_to_pdf(self, html_content: str, pagename: str, page_metadata: Dict[str, str]):
        start = timer()

        abs_dest_path = Path(self._config["outdir"]).joinpath(f"{pagename}.html")
        src_path = Path(f"{pagename}.rst")

        self._options.rst_src_path = Path(self._config["srcdir"]).joinpath(src_path)

        dest_path = abs_dest_path.parent
        if not dest_path.is_dir():
            dest_path.mkdir(parents=True, exist_ok=True)
        self._options.out_dest_path = dest_path

        page_metadata = page_metadata if page_metadata is not None else {}
        pdf_meta: Dict[str, Any] = get_pdf_metadata(page_metadata)
        build_pdf_document = pdf_meta.get("build")

        if self._options.debug and self._options.debug_target is not None:
            # Debugging only the debug target file
            path_filter = URLFilter(self._options, self._config)
            debug_target_file = path_filter(pathname=str(self._options.debug_target))
            doc_src_path = path_filter(pathname=str(self._options.rst_src_path))
            if doc_src_path == debug_target_file:
                build_pdf_document = True
            else:
                build_pdf_document = False

        if build_pdf_document and Path(self._config["srcdir"]).joinpath(src_path).exists():
            self._options.body_title = h1_title_tag(html_content, pdf_meta.get("title"))

            file_name = pdf_meta.get("filename") or pdf_meta.get("title") or self._options.body_title or None
            if file_name is None:
                file_name = str(pagename).split("/")[-1]
                show(
                    context=f"You must set the filename metadata in {pagename}.rst so we can use in the PDF document. "
                    f"The source filename is used as fallback.",
                    error=True,
                )

            # Generate a secure filename
            file_name = secure_filename(file_name)
            base_url = dest_path.joinpath(file_name).as_uri()
            pdf_file = file_name + ".pdf"

            try:
                show(context=f"Converting {src_path} to {pdf_file}")
                self.renderer.write_pdf(
                    html_content,
                    base_url,
                    dest_path.joinpath(pdf_file),
                    pdf_metadata=pdf_meta,
                )

                html_content = self.renderer.add_link(html_content, pdf_file)
                self.pdf_num_files += 1
            except Exception as e:
                self.num_errors += 1
                raise PDFGenerateException(f"Error converting {src_path}. Reason: {e}")
        else:
            show(context=f"Skipped: PDF conversion for {src_path}")

        end = timer()
        self.total_time += end - start
        return html_content


class PDFGenerateException(Exception):
    pass



"""Logic for interacting with sphinx-build."""

import shlex
import subprocess
import sys
from typing import List, Optional, Union

from colorama import Fore, Style

SPHINX_BUILD_OPTIONS = (
    ("b", "builder"),
    ("a", None),
    ("E", None),
    ("d", "path"),
    ("j", "N"),
    ("c", "path"),
    ("C", None),
    ("D", "setting=value"),
    ("t", "tag"),
    ("A", "name=value"),
    ("n", None),
    ("v", None),
    ("q", None),
    ("Q", None),
    ("w", "file"),
    ("W", None),
    ("T", None),
    ("N", None),
    ("P", None),
)


def _log(text: str, *, colour: str) -> None:
    print(f"{Fore.GREEN}[sphinx-pdf-generate] {colour}{text}{Style.RESET_ALL}")


def show(*, context: Optional[str] = None, command: Union[list, tuple, None] = None, error: Optional[bool] = None):
    """Show context and command-to-be-executed, with nice formatting and colors."""
    if context is not None:
        _log(context, colour=Fore.CYAN)
    if command is not None:
        assert isinstance(command, (list, tuple))
        _log("> " + " ".join(shlex.quote(s) for s in command), colour=Fore.MAGENTA)
    if error is not None:
        _log(context, colour=Fore.RED)


def get_builder(sphinx_args: List[str]) -> int:
    """Prepare the function that calls sphinx."""
    sphinx_command = [sys.executable, "-m", "sphinx"] + sphinx_args

    def build() -> int:
        """Generate the documentation using ``sphinx``."""

        try:
            remove_dir_command = ["rm", "-fr", sphinx_args[-1]]
            subprocess.run(remove_dir_command, check=True)
            show(command=["sphinx-build"] + sphinx_args)
            sphinx_build = subprocess.run(sphinx_command, check=True)
            return sphinx_build.returncode
        except subprocess.CalledProcessError as e:
            print(f"Command exited with exit code: {e.returncode}")
            return e.returncode

    return build()



from typing import Any, Dict


class _FilterBase:
    def __init__(self, options: object, config: Dict[str, Any]):
        self.__options = options
        self.__config = config

    @property
    def options(self) -> object:
        return self.__options

    @property
    def config(self) -> Dict[str, Any]:
        return self.__config

    def __call__(self, *args: Any) -> Exception:
        raise Exception("must be overridden")



# import os
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List

import jinja2

from .filters.datetime import strftime, strptime
from .filters.url import URLFilter


class Template:

    """Pickups key-value from `mkdocs.yml`"""

    __KEYS = [
        "author",
        "author_logo",
        "copyright",
        "disclaimer",
        "cover_title",
        "cover_subtitle",
        "site_url",
        "project",
        "version",
        "release",
    ]

    def __init__(self, options: object, config: Dict[str, Any]):
        self._options = options
        self._config = config

        self._keywords = None
        self._jinja_env = None

    @property
    def _env(self) -> jinja2.Environment:
        def generate():
            base_path = Path(Path(__file__).parent).resolve()
            template_paths = []
            docs_src_dir = Path(self._options.srcdir).resolve()
            # Include the template path specified under the `pdfgen_custom_template_path` configuration under conf.py
            custom_template_path = Path(self._options.custom_template_path)
            if not custom_template_path.is_absolute():
                custom_template_path = docs_src_dir.joinpath(self._options.custom_template_path)
            if custom_template_path.is_dir():
                template_paths.append(custom_template_path)

            # Include all templates path listed under the Sphinx `templates_path` configuration under conf.py
            sphinx_templates_paths: List[str] = self._config["templates_path"]
            for _templates_path in sphinx_templates_paths:
                sphinx_template_path = Path(_templates_path)
                if not sphinx_template_path.is_absolute():
                    sphinx_template_path = docs_src_dir.joinpath(sphinx_template_path)
                if sphinx_template_path.is_dir():
                    template_paths.append(sphinx_template_path)

            template_paths.append(base_path.joinpath("."))  # Include the extensions default templates folder path

            template_paths = list(set(template_paths))
            file_loader = jinja2.FileSystemLoader(template_paths)
            logging_undefined = jinja2.make_logging_undefined(logger=self._options.logger, base=jinja2.Undefined)
            env = jinja2.Environment(
                loader=file_loader,
                undefined=logging_undefined,
                lstrip_blocks=True,
                trim_blocks=True,
                autoescape=True,
            )

            env.filters["strptime"] = strptime
            env.filters["strftime"] = strftime
            env.filters["to_url"] = URLFilter(self._options, self._config)

            return env

        if not self._jinja_env:
            self._jinja_env = generate()
        return self._jinja_env

    @property
    def keywords(self) -> dict:
        """Keywords to pass when rendering the template."""

        import html

        def unescape_html_in_list(values: list) -> list:
            new_values = []
            for v in values:
                if isinstance(v, str):
                    new_values.append(html.unescape(v))
                elif isinstance(v, list):
                    new_values.append(unescape_html_in_list(v))
                elif isinstance(v, dict):
                    unescape_html(v)
                    new_values.append(v)
                else:
                    new_values.append(v)
            return new_values

        def unescape_html(variables: dict):
            for k, v in variables.items():
                if isinstance(v, str):
                    variables[k] = html.unescape(v)
                elif isinstance(v, list):
                    variables[k] = unescape_html_in_list(v)
                elif isinstance(v, dict):
                    unescape_html(v)

        def build_keywords():
            keywords = {}

            for key in self.__KEYS:
                if hasattr(self._options, key):
                    keywords[key] = getattr(self._options, key)
                elif key in self._config:
                    keywords[key] = self._config[key]

            unescape_html(keywords)

            keywords["now"] = datetime.now()

            if self._options.verbose:
                from pprint import pformat

                self._options.logger.info("Template variables:")
                for line in pformat(keywords).split("\n"):
                    self._options.logger.info("  " + line)

            return keywords

        if not self._keywords:
            self._keywords = build_keywords()

        return self._keywords

    def select(self, names: [str], parent=None, globals=None) -> jinja2.Template:
        """Find and load a template by names of given."""

        real_names = []
        for name in names:
            for ext in [".html.j2", ".html.jinja2", ".html", ".htm"]:
                real_names.append(name + ext)

        return self._env.select_template(real_names, parent=parent, globals=globals)



import re
from typing import Any, Dict, Optional

from bs4 import BeautifulSoup, PageElement, Tag

from .options import Options
from .templates.filters.url import URLFilter


def make_cover(soup: PageElement, options: Options, config: Dict[str, Any], pdf_metadata: Optional[Dict] = None):
    """Generate a cover pages.

    Arguments:
        soup {BeautifulSoup} -- target element.
        options {Options} -- the project options.
    """

    if options.cover:
        _make_cover(soup, options, config, pdf_metadata)


def _make_cover(soup: PageElement, options: Options, config: Dict[str, Any], pdf_metadata: Optional[Dict] = None):
    try:
        keywords = options.template.keywords
        keywords["site_url"] = re.sub(r"http://|https://", "", keywords["site_url"])
        # Set cover title
        keywords["cover_title"] = pdf_metadata.get("title") or options.body_title or keywords["cover_title"]
        # Set cover image
        document_type: str = pdf_metadata.get("type", "Documentation")
        path_filter = URLFilter(options, config)
        if options.cover_images is not None:
            cover_images = {
                str(img_k).lower(): path_filter(pathname=str(img_v)) for img_k, img_v in options.cover_images.items()
            }
            keywords["cover_image"] = cover_images.get(document_type.lower()) or cover_images.get("default")
        # Set cover sub_title
        keywords["cover_subtitle"] = (
            pdf_metadata.get("subtitle") or document_type.capitalize() or keywords["cover_subtitle"]
        )
        keywords["revision"] = pdf_metadata.get("revision") or None
        # Populate local options into template keywords
        keywords.update(pdf_metadata)

        # Select cover template
        cover_template_files = [document_type.lower(), "cover", "default_cover"]
        template = options.template.select(cover_template_files)

        options.logger.info(f'Generate cover page for PDF document using "{template.name}" template.')

        def str_to_bs4(html_like_str: str) -> Tag:
            html_soup = BeautifulSoup(html_like_str, "html5lib")
            html_tags = html_soup.body.find()
            return html_tags

        cover_template = str(template.render(keywords))
        cover_html = str_to_bs4(cover_template)

        soup.body.insert(0, cover_html)
    except Exception as e:
        options.logger.error("Failed to generate the cover page: %s", e)



import logging
from pathlib import Path
from typing import Any, Dict, Optional

from .templates.filters.url import URLFilter
from .templates.template import Template


class Options:
    def __init__(self, config: Dict[str, Any], logger: logging):
        self.verbose: bool = config["verbose"]
        self.debug: bool = config["debug"]
        self.debug_target: Optional[str] = config["debug_target"]
        self.srcdir: Path = Path(config["srcdir"])
        self.outdir: Path = Path(config["outdir"])
        self._src_path: Optional[Path] = None
        self._dest_path: Optional[Path] = None

        self.site_url: str = config["site_url"]
        self.theme_name: str = config["theme_name"]

        # Author, Copyright, Disclaimer
        self._author: Optional[str] = config["author"]
        self._copyright: Optional[str] = config["copyright"]
        self._disclaimer: Optional[str] = config["disclaimer"]

        # Individual document type cover
        self._cover_images: Optional[Dict] = config["cover_images"]

        # Cover
        self.cover: bool = config["cover"]
        self._cover_title: Optional[str] = config["cover_title"]
        self._cover_subtitle: Optional[str] = config["cover_subtitle"]

        # path to custom template 'cover.html' and 'custom.css'
        self.custom_template_path: str = config["custom_template_path"]
        self.custom_css_path: Optional[str] = config["custom_css_path"]

        # TOC and Chapter heading
        self.toc: bool = config["toc"]
        self.toc_title: str = config["toc_title"]
        self.toc_level: int = config["toc_level"]
        self.toc_ordering: bool = config["toc_numbering"]

        # H1 Title of the document
        self._body_title: str = ""

        # Template handler(Jinja2 wrapper)
        self._template = Template(self, config)

        # Author Logo
        logo_path_filter = URLFilter(self, config)
        self.author_logo: Optional[str] = config["author_logo"]
        if isinstance(self.author_logo, str):
            self.author_logo = logo_path_filter(self.author_logo)

        # Custom Theme and User Plugin
        self.theme_handler_path: Optional[str] = config["theme_handler_path"]
        self.user_plugin_handler_path: Optional[str] = config["plugin_handler_path"]

        # for system
        self._logger = logger

    @property
    def body_title(self) -> str:
        return self._body_title

    @body_title.setter
    def body_title(self, text):
        self._body_title = text

    @property
    def author(self) -> str:
        return self._author

    @property
    def copyright(self) -> str:
        return self._copyright

    @property
    def disclaimer(self) -> str:
        return self._disclaimer

    @property
    def cover_title(self) -> str:
        return self._cover_title

    @property
    def cover_subtitle(self) -> str:
        return self._cover_subtitle

    @property
    def cover_images(self) -> Dict:
        return self._cover_images

    @property
    def logger(self) -> logging:
        return self._logger

    @property
    def template(self) -> Template:
        return self._template

    @property
    def rst_src_path(self) -> Path:
        return self._src_path

    @rst_src_path.setter
    def rst_src_path(self, input_path):
        self._src_path = input_path

    @property
    def out_dest_path(self) -> Path:
        return self._dest_path

    @out_dest_path.setter
    def out_dest_path(self, input_path):
        self._dest_path = input_path

    def debug_dir(self) -> Path:
        if self.debug:
            docs_out_dir = Path(self.outdir).parent.resolve()
            debug_folder_path = docs_out_dir.joinpath("pdf_html_debug")
            if not debug_folder_path.is_dir():
                debug_folder_path.mkdir(parents=True, exist_ok=True)
            return debug_folder_path



from bs4 import BeautifulSoup, NavigableString, PageElement, Tag

from .options import Options


def make_toc(soup: BeautifulSoup, options: Options):
    """Generate a toc tree.

    Arguments:
        soup {BeautifulSoup} -- target element.
        options {Options} -- the project options.
    """

    if options.toc:
        _make_indexes(soup, options)


def _make_indexes(soup: BeautifulSoup, options: Options) -> None:
    """Generate ordered chapter number and TOC of document.

    Arguments:
        soup {BeautifulSoup} -- HTML content.
        options {Options} -- The options of this sequence.
    """

    # Step 1: (re)ordered headings
    if options.toc_ordering:
        _inject_heading_order(soup, options)

    # Step 2: generate toc page
    level = options.toc_level
    if level < 1 or level > 6:
        return

    options.logger.info(f"Generate table of contents up to heading level {level} for PDF document.")

    h1li = None
    h2ul = h2li = h3ul = h3li = h4ul = h4li = h5ul = h5li = h6ul = None
    # exclude_lv2 = exclude_lv3 = False

    def makelink(heading: PageElement) -> PageElement:
        li = soup.new_tag("li")
        ref = heading.get("id", "")
        if ref == "":
            ref = heading.parent.get("id")  # support for all sphinx themes
        prefix = heading.get("data-numbering")
        a = (
            soup.new_tag("a", href=f"#{ref}", attrs={"data-numbering": prefix})
            if prefix is not None
            else soup.new_tag("a", href=f"#{ref}")
        )
        for el in heading.contents:
            if el.name == "a":
                a.append(el.contents[0].replace("¶", ""))
            else:
                a.append(_clone_element(el))
        li.append(a)
        options.logger.debug(f"| [{heading.get_text(separator=' ')}]({ref})")
        return li

    toc = soup.new_tag("article", id="doc-toc")
    title = soup.new_tag("h1")
    title.append(soup.new_string(options.toc_title))
    toc.append(title)

    h1ul = soup.new_tag("ul")
    toc.append(h1ul)

    headings = soup.find_all(["h1", "h2", "h3", "h4", "h5", "h6"])
    for h in headings:
        if h.name == "h1":
            h1li = makelink(h)
            h1ul.append(h1li)
            h2ul = h2li = h3ul = h3li = h4ul = h4li = h5ul = h5li = h6ul = None

            # exclude_lv2 = _is_exclude(h.get("id", None), options)

        elif h.name == "h2" and level >= 2:
            if not h2ul:
                h2ul = soup.new_tag("ul")
                h1li.append(h2ul)
            h2li = makelink(h)
            h2ul.append(h2li)
            h3ul = h3li = h4ul = h4li = h5ul = h5li = h6ul = None

            # exclude_lv3 = _is_exclude(h.get("id", None), options)

        elif h.name == "h3" and level >= 3:
            if not h2li:
                continue
            if not h3ul:
                h3ul = soup.new_tag("ul")
                h2li.append(h3ul)
            h3li = makelink(h)
            h3ul.append(h3li)
            h4ul = h4li = h5ul = h5li = h6ul = None

        elif h.name == "h4" and level >= 4:
            if not h3li:
                continue
            if not h4ul:
                h4ul = soup.new_tag("ul")
                h3li.append(h4ul)
            h4li = makelink(h)
            h4ul.append(h4li)
            h5ul = h5li = h6ul = None

        elif h.name == "h5" and level >= 5:
            if not h4li:
                continue
            if not h5ul:
                h5ul = soup.new_tag("ul")
                h4li.append(h5ul)
            h5li = makelink(h)
            h5ul.append(h5li)
            h6ul = None

        elif h.name == "h6" and level >= 6:
            if not h5li:
                continue
            if not h6ul:
                h6ul = soup.new_tag("ul")
                h5li.append(h6ul)
            h6li = makelink(h)
            h6ul.append(h6li)

        else:
            continue

    soup.body.insert(0, toc)


def _inject_heading_order(soup: BeautifulSoup, options: Options):
    level = options.toc_level
    if level < 1 or level > 6:
        return

    options.logger.debug(f"Number headings up to level {level}.")

    h1n = h2n = h3n = h4n = h5n = h6n = 0
    # exclude_lv2 = exclude_lv3 = False

    headings = soup.find_all(["h1", "h2", "h3", "h4", "h5", "h6"])
    for h in headings:
        if h.name == "h1":
            h1n += 1
            h2n = h3n = h4n = h5n = h6n = 0
            prefix = f"{h1n}. "

        # exclude_lv2 = _is_exclude(h.get("id", None), options)

        elif h.name == "h2" and level >= 2:
            h2n += 1
            h3n = h4n = h5n = h6n = 0
            prefix = f"{h1n}.{h2n}. "

            # exclude_lv3 = _is_exclude(h.get("id", None), options)

        elif h.name == "h3" and level >= 3:
            h3n += 1
            h4n = h5n = h6n = 0
            prefix = f"{h1n}.{h2n}.{h3n}. "

        elif h.name == "h4" and level >= 4:
            h4n += 1
            h5n = h6n = 0
            prefix = f"{h1n}.{h2n}.{h3n}.{h4n}. "

        elif h.name == "h5" and level >= 5:
            h5n += 1
            h6n = 0
            prefix = f"{h1n}.{h2n}.{h3n}.{h4n}.{h5n}. "

        elif h.name == "h6" and level >= 6:
            h6n += 1
            prefix = f"{h1n}.{h2n}.{h3n}.{h4n}.{h5n}.{h6n}. "

        else:
            continue

        h["data-numbering"] = prefix


def _clone_element(el: Tag) -> Tag:
    if isinstance(el, NavigableString):
        return type(el)(el)

    copy = Tag(None, el.builder, el.name, el.namespace, el.nsprefix)
    # work around bug where there is no builder set
    # https://bugs.launchpad.net/beautifulsoup/+bug/1307471
    copy.attrs = dict(el.attrs)
    for attr in ("can_be_empty_element", "hidden"):
        setattr(copy, attr, getattr(el, attr))
    for child in el.contents:
        copy.append(_clone_element(child))
    return copy



"""Main implementation."""

import argparse
import json
import os
from pathlib import Path
from typing import Dict

import colorama

from .build import SPHINX_BUILD_OPTIONS, get_builder, show
from .pdf_generate import PdfGeneratePlugin
from .version import __version__

GLOBAL_OPTIONS = {}.update(
    verbose=False,
    site_url="http://127.0.0.1:8000",
    debug=False,
    debug_target=None,
    author=None,
    author_logo=None,
    copyright=None,
    disclaimer=None,
    cover=True,
    cover_title=None,
    cover_subtitle=None,
    custom_template_path="_templates",
    theme_handler_path=None,
    plugin_handler_path=None,
    custom_css_path=None,
    toc=True,
    toc_numbering=True,
    toc_title="Table of Contents",
    toc_level=4,
    cover_images=None,
    theme_name="alabaster",
    templates_path=[],
)


def _get_build_args(args):
    build_args = []
    for arg, meta in SPHINX_BUILD_OPTIONS:
        val = getattr(args, arg)
        if not val:
            continue
        opt = f"-{arg}"
        if meta is None:
            build_args.extend([opt] * val)
        else:
            for v in val:
                build_args.extend([opt, v])

    build_args.extend([os.path.realpath(args.sourcedir), os.path.realpath(args.outdir)])
    return build_args


def get_parser():
    """Get the application's argument parser.

    Note: this also handles SPHINX_BUILD_OPTIONS, which later get forwarded to
    sphinx-build as-is.
    """

    class RawTextArgumentDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter):
        pass

    parser = argparse.ArgumentParser(
        formatter_class=RawTextArgumentDefaultsHelpFormatter,
        description="Build PDF documents from Sphinx HTML build files.",
    )

    parser.add_argument("--version", action="version", version=f"sphinx-pdf-generate {__version__}")

    sphinx_arguments = ", ".join(f"-{arg}" if meta is None else f"-{arg}={meta}" for arg, meta in SPHINX_BUILD_OPTIONS)
    sphinx_parser = parser.add_argument_group(
        "Sphinx's arguments",
        (
            "The following arguments are forwarded as-is to Sphinx. Please look at "
            f"`sphinx-build --help` for more information.\n  {sphinx_arguments}"
        ),
    )

    for arg, meta in SPHINX_BUILD_OPTIONS:
        if meta is None:
            sphinx_parser.add_argument(f"-{arg}", action="count", help=argparse.SUPPRESS)
        else:
            sphinx_parser.add_argument(
                f"-{arg}",
                action="append",
                help=argparse.SUPPRESS,
                metavar=meta,
            )

    parser.add_argument("sourcedir", help="source directory")
    parser.add_argument("outdir", help="output directory for built documentation")
    return parser


def main() -> None:
    """Actual application logic."""
    colorama.init()

    parser = get_parser()
    args = parser.parse_args()

    show(context="Starting PDF Build Process")

    srcdir = os.path.realpath(args.sourcedir)
    outdir = os.path.realpath(args.outdir)

    if not os.path.exists(srcdir):
        raise PDFGenerateException(f"The source directory {srcdir} does not exist.")
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    build_args = _get_build_args(args)
    builder = get_builder(build_args)

    if builder == 0:
        # Load configuration
        metadata_options_file = Path(outdir).joinpath("pdf_metadata.json")
        if not os.path.exists(metadata_options_file):
            raise PDFGenerateException(
                f"The The PDF metadata file, 'pdf_metadata.json', not found in the output directory. Check: {outdir}"
            )

        with open(metadata_options_file) as json_file:
            load_options: Dict[str, Dict] = json.load(json_file)

        global_config = load_options["GLOBAL_OPTIONS"] if "GLOBAL_OPTIONS" in load_options else GLOBAL_OPTIONS
        local_config = load_options.get("LOCAL_OPTIONS")

        if not local_config:
            raise PDFGenerateException(
                f"The PDF metadata file does not contain information about the local options. "
                f"Check the file for more information: {metadata_options_file}."
            )

        global_config.update(outdir=outdir, srcdir=srcdir)
        pdf_generator = PdfGeneratePlugin()
        pdf_generator.on_config(global_config)

        for html_pagename, html_metadata in local_config.items():
            html_page_path = Path(outdir).joinpath(f"{html_pagename}.html")
            html_page_content = html_page_path.read_text(encoding="utf-8")
            new_html_page_content = pdf_generator.convert_page_to_pdf(
                html_content=html_page_content, pagename=html_pagename, page_metadata=html_metadata
            )
            html_page_path.write_text(data=new_html_page_content, encoding="utf-8")

        show(context=f"Converting {pdf_generator.pdf_num_files} file(s) to PDF took {pdf_generator.total_time:.1f}s")

        if pdf_generator.num_errors > 0:
            show(context=f"{pdf_generator.num_errors} conversion errors occurred (see above)", error=True)
    else:
        show(context="Sphinx build was unsuccessful. No PDF files were generated.", error=True)


class PDFGenerateException(Exception):
    pass
